--imports
local stringx = require("pl.stringx")
local Reader = require("dogma.lex._.Reader")
local ProcessedList = require("dogma.lex._.ProcessedList")
local AdvancedList = require("dogma.lex._.AdvancedList")
local Eol = require("dogma.lex._.Eol")
local Annotation = require("dogma.lex._.Annotation")
local Comment = require("dogma.lex._.Comment")
local Directive = require("dogma.lex._.Directive")
local Name = require("dogma.lex._.Name")
local Keyword = require("dogma.lex._.Keyword")
local Literal = require("dogma.lex._.Literal")
local LiteralType = require("dogma.lex.LiteralType")
local Re = require("dogma.lex._.Re")
local Symbol = require("dogma.lex._.Symbol")
local TokenType = require("dogma.lex.TokenType")

--A lexer os scanner.
local Lexer = {}
Lexer.__index = Lexer
package.loaded[...] = Lexer

--Constructor.
--
--@param props:object Lexer properties.
function Lexer.new(props)
  local self

  --(1) arguments
  if not props then props = {} end

  --(2) create
  self = setmetatable({
    _ = {
      comments = not not props.comments,
      paused = {}
    }
  }, Lexer)

  --(3) return
  return self
end

--Raises an error.
--
--@param msg:string
function Lexer:_raise(msg, tok, col)
  local ln

  if type(tok) == "table" then
    ln, col = tok.line, tok.col
  else
    ln = tok
  end

  if ln ~= nil then
    error(string.format("%s (%s, %s): %s", self._.file, ln, col, msg))
  else
    error(string.format("%s: %s", self._.file, msg))
  end
end

--Scan a text.
--
--@param text:string  Text to analyze.
--@param file?:string File path.
--@param pause?:bool  Pause the current scan?
--
--@return self
function Lexer:scan(text, file, pause)
  --(1) arguments
  if not text then self:_raise("text expected.") end

  --(2) init
  if pause then
    table.insert(self._.paused, {
      reader = self._.reader,
      file = self._.file,
      processed = self._.processed,
      token = self._.token,
      advanced = self._.advanced
    })
  else
    self._paused = {}
  end

  self._.reader = Reader.new(text)
  self._.file = file or "anonymous code"
  self._.processed = ProcessedList.new(3)
  self._.token = nil
  self._.advanced = AdvancedList.new(3)

  --(3) return
  return self
end

--Check whether it has some scan paused.
function Lexer:_hasPaused()
  return #self._.paused > 0
end

--Resume a pause.
function Lexer:_resume()
  local i = table.remove(self._.paused)

  self._.reader = i.reader
  self._file = i.file
  self._.processed = i.processed
  self._.token = i.token
  self._.advanced = i.advanced

  return self
end

--Return the last token that has been read.
--
--@return Token
function Lexer:_getLastReadToken()
  return self._.token
end

--Return the current token.
--
--@param Token
function Lexer:current()
  return self._.token
end

--Scan the next token.
--
--@overload
--@return Token
--
--@overload
--@param typ:TokenType  Token type to read.
--@param val:string     Token value to read.
--@return Token
function Lexer:next(typ, val)
  local reader = self._.reader
  local tok

  --(1) get token to return
  if self:_hasTokenToShift() then
    self:_shift()
  else
    local ch
    local function nop() end

    --get next char
    while true do
      ch = reader:next()

      if ch == nil then
        break
      else
        if ch.char == " " or ch.char == "\t" then
          nop()
        elseif ch.char == "\\" then
          ch = reader:next()

          if ch ~= nil and ch.char == "\n" then
            nop()
          else
            self:_raise("end of line expected.", ch)
          end
        else
          break
        end
      end
    end

    --end of the file and paused?
    if ch == nil and self:_hasPaused() then
      self:_resume()
      return self:next(typ, val)
    end

    --scan next token
    if ch ~= nil then
      ch = ch.char

      --analyze token
      reader:unshift()

      if ch == "\n" then
        tok = self:_scanEol()
      elseif ch == "@" then
        tok = self:_scanAnnotation()
      elseif ch == "_" or stringx.isalpha(ch) then
        tok = self:_scanId()
      elseif ch == "'" then
        tok = self:_scanName()
      elseif stringx.isdigit(ch) then
        tok = self:_scanLiteralNumber()
      elseif ch == "$" then
        reader:next() --$

        local aux = reader:next()
        reader:unshift()
        reader:unshift()

        if aux.char == '"' then
          tok = self:_scanLiteralString()
        else
          tok = self:_scanSymbol()
        end
      elseif ch == '"' then
        tok = self:_scanLiteralString()
      elseif ch == '`' then
        tok = self:_scanLiteralRe()
      elseif ch == "/" then
        local _, adv = reader:next(), reader:next()
        reader:unshift()
        reader:unshift()

        if adv ~= nil and adv.char == "*" then
          tok = self:_scanDocComment()

          if not self._.comments then
            return self:next()
          end
        else
          tok = self:_scanSymbol()
        end
      elseif ch == "#" then
        local _, adv1, adv2 = reader:next(), reader:next(), reader:next()

        reader:unshift()
        reader:unshift()
        reader:unshift()

        if (adv1 and adv1.char == "!") or (adv1 and adv2 and adv1.char == "[" and adv2.char == "!") then
          tok = self:_scanDirective()
        else
          tok = self:_scanComment()

          if not self._.comments then
            return self:next()
          end
        end
      else
        tok = self:_scanSymbol()
      end

      self:_backUpToken()
      self._.token = tok
    else
      self:_backUpToken()
      self._.token = nil
    end
  end

  --(2) check
  if typ then
    tok = self._.token

    if tok == nil then
      self:_raise(string.format("'%s' expected at the end of code.", val))
    end

    if tok.type ~= typ or (val ~= nil and tok.value ~= val) then
      if typ == TokenType.EOL then
        self:_raise("end of line expected.", tok)
      elseif typ == TokenType.NAME then
        self:_raise("name expected.", tok)
      elseif typ == TokenType.LITERAL then
        self:_raise("literal expected.", tok)
      else
        self:_raise(string.format("'%s' expected.", val), tok)
      end
    end
  end

  --(3) return
  if self._.token and self._.token:isDirective("include") then
    return self:_include(self._.token.value:gsub("^include ", ""), typ, val)
  else
    return self._.token
  end
end

--Process a #!include directive.
function Lexer:_include(filename)
  local file = require("pl.file")
  local path = require("pl.path")

  --(1) pre
  if self._.file == "anonymous code" then
    self:_raise("anonymous code doesn't accept #!include.", self._.token)
  end

  --(2) get file path
  filename = path.join(path.dirname(self._.file), filename .. ".dogi")

  --(3) scan
  local content = file.read(filename)

  if not content then
    self:_raise(string.format("file not found: %s.", filename), self._.token)
  end

  return self:scan(content, filename, true):next()
end

--Insert a text and process.
--Used by parser when template.
function Lexer:insert(content, filename)
  return self:scan(content, filename, true)
end

--Return the next token that must be an end of line.
--
--@return Eol
function Lexer:nextEol()
  return self:next(TokenType.EOL)
end

function Lexer:nextNewline()
  return self:next(TokenType.EOL)
end

--Skip the next token if it is an end of line.
--
--@return Eol
function Lexer:skipEol()
  local tok = self:advance()

  if tok and tok:isEol() then
    return self:nextEol()
  end
end

--Return the next token that must be an id: keyword or name.
--
--@return Id
function Lexer:nextId()
  local tok = self:advance()

  if tok.type ~= TokenType.KEYWORD and tok.type ~= TokenType.NAME then
    self:_raise("id expected.", tok)
  end

  return self:next()
end

--Return the next token that must be a name.
--
--@return Name
function Lexer:nextName(id)
  local tok = self:next(TokenType.NAME)

  if id and tok.value ~= id then
    self:_raise(string.format("name expected: '%s'. Received: '%s'.", id, tok.value), tok)
  end

  return tok
end

--Return the next token if name.
--
--@return Name
function Lexer:skipName()
  local tok = self:advance()

  if tok and tok:isName() then
    return self:nextName()
  end
end

--Return the next token that must be a symbol.
--
--@return Symbol
function Lexer:nextSymbol(val)
  return self:next(TokenType.SYMBOL, val)
end

--Read the next token being it a comma symbol or an end of line.
function Lexer:nextCommaSymbolOrEol()
  local tok = self:skipSymbol(",")

  if not tok then
    tok = self:skipEol()
  end

  if not tok then
    tok = self:next()
    self:_raise("comma symbol or end of line expected.", tok)
  end

  return tok
end

--Read the next token if it is a symbol.
--
--@return Symbol
function Lexer:skipSymbol(val)
  local tok = self:advance()

  if tok and tok:isSymbol(val) then
    return self:nextSymbol(val)
  end
end

--Read the next tokens if these are given symbols.
--
--@return Symbol[]
function Lexer:skipSymbols(val1, val2, val3)
  local tok = self:advance(1)

  if tok and tok:isSymbol(val1) then
    tok = self:advance(2)

    if tok and tok:isSymbol(val2) then
      if val3 ~= nil then
        tok = self:advance(3)

        if tok and tok:isSymbol(val3) then
          return {self:next(), self:next(), self:next()}
        end
      else
        return {self:next(), self:next()}
      end
    end
  end
end

--Return the next token that must be a keyword.
--
--@return Keyword
function Lexer:nextKeyword(val)
  return self:next(TokenType.KEYWORD, val)
end

--Skip the next token if it is the specified keyword.
--
--@return Keyword
function Lexer:skipKeyword(val)
  local tok = self:advance()

  if tok and tok:isKeyword(val) then
    return self:next()
  end
end

--Advance the next token.
--
--@return Token
function Lexer:advance(i)
  local tok

  --(1) arguments
  if i == nil then
    i = 1
  end

  --(2) get token
  if #self._.advanced >= i then
    tok = self._.advanced._.items[i]
  else
    if i == 1 then
      tok = self:next()
      self:unshift()
    elseif i == 2 then
      self:next()
      tok = self:next()
      self:unshift()
      self:unshift()
    elseif i == 3 then
      self:next()
      self:next()
      tok = self:next()
      self:unshift()
      self:unshift()
      self:unshift()
    end
  end

  --(3) return
  return tok
end

--Check whether the lexer has some token to shift.
--
--@return bool
function Lexer:_hasTokenToShift()
  return #self._.advanced > 0
end

--Shift a token: <- processed <- current <- advanced
function Lexer:_shift()
  --(1) pre
  if #self._.advanced == 0 then
    self:_raise("no advanced token to shift.")
  end

  --(2) current to processed
  if self._.token then
    self._.processed:insert(self._.token)
  end

  --(3) advanced to current
  self._.token = self._.advanced:remove()
end

--Unshift a char: processed -> current -> advanced
function Lexer:unshift()
  --(1) pre
  if not self._.token and #self._.processed == 0 then
    self:_raise("no current token to unshift.")
  end

  --(2) current to advanced
  self._.advanced:insert(self._.token)

  --(3) last processed to current
  if #self._.processed == 0 then
    self._.token = nil
  else
    self._.token = self._.processed:remove()
  end
end

--Shift current token to processed tokens: processed <- token.
function Lexer:_backUpToken()
  if self._.token then
    self._.processed:insert(self._.token)
  end
end

--Scan an end of line.
--
--@return Eol
function Lexer:_scanEol()
  local reader = self._.reader
  local ch = reader:next()

  return Eol.new(ch.line, ch.col)
end

--Scan an annotation.
--
--@return Annotation
function Lexer:_scanAnnotation()
  local rdr = self._.reader
  local state, ln, col, val, args
  local State = {
    START = 1,
    VALUE = 2,
    END = 3
  }

  --(1) state machine
  state = State.START

  while state ~= State.END do
    local ch = rdr:next()

    if state == State.START then
      ln, col, val = ch.line, ch.col, ""
      state = State.VALUE
    elseif state == State.VALUE then
      ch = ch.char

      if not (stringx.isalnum(ch) or ch == "_") then
        rdr:unshift()
        state = State.END
      else
        val = val .. ch
      end
    end
  end

  --(2) read argument
  local ch = rdr:next()

  if ch and ch.char == "(" then
    ch = rdr:next()
    if ch.char ~= '"' then
      self:_raise("Literal text expected.", ch.line, ch.col)
    else
      rdr:unshift()
    end

    args = {self:_scanLiteralString().value}
    ch = rdr:next()
    if ch.char ~= ')' then
      self:_raise("')' expected.", ch.line, ch.col)
    end
  else
    rdr:unshift()
  end

  --(3) return
  return Annotation.new(ln, col, val, args)
end

--Scan a directive.
--
--@return Directive
function Lexer:_scanDirective()
  local rdr = self._.reader
  local state, ln, col, val, inline
  local State = {
    START1 = 1,
    START2 = 2,
    START3 = 3,
    VALUE = 4,
    END = 5
  }

  --(1) state machine
  state = State.START1

  while state ~= State.END do
    local ch = rdr:next()

    if state == State.START1 then     --#
      ln, col, val = ch.line, ch.col, ""
      state = State.START2
    elseif state == State.START2 then --[ or !
      if ch.char == "[" then
        inline = true
        state = State.START3
      else
        inline = false
        val = ""
        state = State.VALUE
      end
    elseif state == State.START3 then --! from [
      state = State.VALUE
      val = ""
    elseif state == State.VALUE then
      ch = ch.char

      if inline then
        if ch == "]" then
          state = State.END
        else
          val = val .. ch
        end
      else
        if ch == "\n" then
          state = State.END
        else
          val = val .. ch
        end
      end
    end
  end

  if not (val == "react" or
          val:find("^include ") or
          val:find("^insert ") or
          val:find("^coverage ignore ") or
          val:find("^cov ignore$") or
          val:find("^cov ignore ") or
          val:find("^if [a-zA-Z]+ then$") or
          val:find("^if [a-zA-Z]+ or [a-zA-Z]+ then$") or
          val:find("^if not [a-zA-Z]+ then$") or
          val:find("^/") or
          val:find("^else if ") or
          val == "env" or
          val == "end" or
          val == "else") then
    self:_raise("invalid directive.", ln, col)
  end

  --(2) return
  return Directive.new(ln, col, val)
end

--Scan a doc comment.
--
--@return Comment
function Lexer:_scanDocComment()
  local reader = self._.reader
  local state, ln, col, comm
  local State = {
    START1 = 1,
    START2 = 2,
    TEXT = 3,
    POSSIBLE_END = 4,
    END = 5
  }

  --(1) state machine
  state = State.START1

  while state ~= State.END do
    local ch = reader:next()

    if not ch then
      self:_raise("Doc comment not ended.", ln, col)
    elseif state == State.START1 then
      ln, col, comm = ch.line, ch.col, ""
      state = State.START2
    elseif state == State.START2 then
      state = State.TEXT
    elseif state == State.TEXT then
      ch = ch.char

      if ch == "*" then
        state = State.POSSIBLE_END
      else
        comm = comm .. ch
      end
    elseif state == State.POSSIBLE_END then
      ch = ch.char

      if ch == "/" then
        state = State.END
      elseif ch == "*" then
        comm = comm .. "*"
      else
        comm = comm .. "*" .. ch
        state = State.TEXT
      end
    end
  end

  --(2) return
  return Comment.new(ln, col, comm)
end

--Scan a comment.
--
--@return Comment
function Lexer:_scanComment()
  local reader = self._.reader
  local state, ln, col, comm, multiln
  local State = {
    START = 1,
    MULTILINE = 2,
    TEXT = 3,
    POSSIBLE_END = 4,
    POSSIBLE_DIRECTIVE = 5,
    END = 6
  }

  --(1) state machine
  state = State.START

  while state ~= State.END do
    local ch = reader:next()

    if state == State.START then
      ln, col, comm = ch.line, ch.col, ""
      state = State.MULTILINE
    elseif state == State.MULTILINE then
      ch = ch.char

      if ch == "[" then
        multiln = true
        comm = ""
        state = State.TEXT
      elseif ch == "\n" then
        comm = ""
        state = State.POSSIBLE_END
        multiln = false
      else
        multiln = false
        comm = ch
        state = State.TEXT
      end
    elseif state == State.TEXT then
      ch = ch.char

      if multiln then
        if ch == "]" then
          state = State.END
        else
          comm = comm .. ch
        end
      else
        if ch == "\n" then
          state = State.POSSIBLE_END
        else
          comm = comm .. ch
        end
      end
    elseif state == State.POSSIBLE_END then
      if ch == nil then
        reader:unshift()  --current: nil
        reader:unshift()  --end of line
        state = State.END
      else
        ch = ch.char

        if ch == "#" then
          state = State.POSSIBLE_DIRECTIVE
        else
          reader:unshift()  --current char
          reader:unshift()  --eol
          state = State.END
        end
      end
    else  --POSSIBLE_DIRECTIVE
      ch = ch.char

      if ch == "!" or ch == "[" then
        reader:unshift()  --! or [
        reader:unshift()  --#
        state = State.END
      else
        reader:unshift()  --current for new processing
        comm = comm .. "\n"
        state = State.TEXT
      end
    end
  end

  --(2) return
  return Comment.new(ln, col, comm)
end

--Scan an identifier.
--
--@return Id
function Lexer:_scanId()
  local reader = self._.reader
  local state, ln, col, id
  local State = {
    START = 1,
    MIDDLE = 2,
    END = 3
  }

  --(1) state machine
  state = State.START

  while state ~= State.END do
    local ch = reader:next()

    if state == State.START then
      ln, col = ch.line, ch.col
      id = ch.char
      state = State.MIDDLE
    else
      ch = ch.char

      if ch == "_" or stringx.isalnum(ch) then
        id = id .. ch
      else
        state = State.END
        reader:unshift()
      end
    end
  end

  --(2) return
  local last = self:_getLastReadToken()

  if (last and last.type == TokenType.SYMBOL and (last.value == "." or last.value == ":" or last.value == "?")) or
     not Keyword.is(id) then
    return Name.new(ln, col, id)
  else
    return Keyword.new(ln, col, id)
  end
end

--Scan a name.
--
--@return Name
function Lexer:_scanName()
  local reader = self._.reader
  local state, ln, col, id, ch
  local State = {
    START = 1,
    ID = 2,
    END = 3
  }

  --(1) state machine
  state = State.START
  ch = reader:next()  --'
  ln, col = ch.line, ch.col

  while state ~= State.END do
    ch = reader:next()

    if state == State.START then
      id = ch.char
      state = State.ID
    else --State.ID
      ch = ch.char

      if ch == "'" then
        state = State.END
      else
        id = id .. ch
      end
    end
  end

  --(2) return
  return Name.new(ln, col, id)
end

--Scan a symbol.
--
--@return Symbol
function Lexer:_scanSymbol()
  local reader = self._.reader
  local state, ln, col, sym
  local State = {
    START = 1,
    MIDDLE = 2,
    END = 3
  }
  local SYMBOLS = {
    ["+"] = true,
    ["+="] = true,
    ["-"] = true,
    ["-="] = true,
    ["->"] = true,
    ["*"] = true,
    ["*="] = true,
    ["**"] = true,
    ["**="] = true,
    ["/"] = true,
    ["/="] = true,
    ["/>"] = true,
    ["%"] = true,
    ["%="] = true,
    ["="] = true,
    ["=="] = true,
    ["==="] = true,
    ["=>"] = true,
    ["=~"] = true,
    ["!"] = true,
    ["!="] = true,
    ["!=="] = true,
    ["!~"] = true,
    ["<"] = true,
    ["<<"] = true,
    ["<<<"] = true,
    ["<<="] = true,
    ["<="] = true,
    ["</"] = true,
    [">"] = true,
    [">>"] = true,
    [">>>"] = true,
    [">>="] = true,
    [">="] = true,
    ["^"] = true,
    ["^="] = true,
    ["~"] = true,
    ["~="] = true,
    ["("] = true,
    [")"] = true,
    ["["] = true,
    ["]"] = true,
    ["{"] = true,
    ["}"] = true,
    [";"] = true,
    [":"] = true,
    ["::"] = true,
    ["::="] = true,
    [":="] = true,
    [":>"] = true,
    [":~"] = true,
    ["."] = true,
    [".."] = true,
    ["..."] = true,
    [".="] = true,
    ["?"] = true,
    ["?="] = true,
    ["&"] = true,
    ["&="] = true,
    ["&&"] = true,
    ["|"] = true,
    ["|="] = true,
    ["||"] = true,
    [","] = true,
    ["$"] = true
  }

  --(1) state machine
  state = State.START

  while state ~= State.END do
    local ch = reader:next()

    if state == State.START then
      if SYMBOLS[ch.char] then
        ln, col, sym = ch.line, ch.col, ch.char
        state = State.MIDDLE
      else
        reader:unshift()
        state = State.END
      end
    else
      if SYMBOLS[sym .. ch.char] then
        sym = sym .. ch.char
      else
        reader:unshift()
        state = State.END
      end
    end
  end

  --(2) return
  if sym == nil then
    self:_raise("invalid symbol.", reader._.line, reader._.col - 1)
  end

  return Symbol.new(ln, col, sym)
end

--Scan a literal hnumber.
--
--@return Literal
function Lexer:_scanLiteralNumber()
  local reader = self._.reader
  local state, ln, col, lit
  local State = {
    START = 1,
    INT = 2,
    DOT = 3,
    DECIMAL = 4,
    END = 5
  }

  --(1) state machine
  state = State.START

  while state ~= State.END do
    local ch = reader:next()

    if state == State.START then
      ln, col, lit = ch.line, ch.col, ch.char
      state = State.INT
    elseif state == State.INT then
      ch = ch.char

      if stringx.isdigit(ch) then
        lit = lit .. ch
      elseif ch == " " then
        lit = lit --nothing to do
      elseif ch == "." then
        state = State.DOT
      else
        reader:unshift()
        state = State.END
      end
    elseif state == State.DOT then
      ch = ch.char

      if stringx.isdigit(ch) then
        lit = lit .. "." .. ch
        state = State.DECIMAL
      else
        reader:unshift() --current
        reader:unshift() --dot
        state = State.END
      end
    else  --DECIMAL
      ch = ch.char

      if stringx.isdigit(ch) then
        lit = lit .. ch
      else
        reader:unshift()
        state = State.END
      end
    end
  end

  --(2) return
  return Literal.new(ln, col, LiteralType.NUMBER, tonumber(lit))
end

--Scan a literal string.
--
--@return Literal
function Lexer:_scanLiteralString()
  local reader = self._.reader
  local state, ln, col, lit, tmpl
  local State = {
    START = 1,
    START1 = 2,
    START2 = 3,
    TEXT1 = 4,
    TEXT3 = 5,
    END1 = 6,
    END2 = 7,
    END = 8,
    TEXT3_START = 9
  }

  local function char(ch)
    ch = ch.char

    if #ch > 1 then
      if ch == "\\\\" then
        ch = "\\"
      elseif ch == "\\\"" then
        ch = "\""
      elseif ch == "\\n" then
        ch = "\n"
      elseif ch == "\\t" then
        ch = "\t"
      end
    end

    return ch
  end

  --(1) template?
  local start = reader:next()
  ln, col, lit = start.line, start.col, ""

  if start.char == "$" then
    tmpl = true
  else
    reader:unshift()
    tmpl = false
  end

  --(2) state machine
  state = State.START

  while state ~= State.END do
    local ch = reader:next()

    if ch and ch.char == "\\" then
      ch = reader:next()
      if ch then
        ch.char = "\\" .. ch.char
      end
    end

    ::START::if state == State.START then
      state = State.START1
    elseif state == State.START1 then
      if ch.char == '"' then
        state = State.START2
      else
        lit = lit .. char(ch)
        state = State.TEXT1
      end
    elseif state == State.START2 then
      if ch.char == '"' then
        state = State.TEXT3_START
      else
        reader:unshift()
        state = State.END
      end
    elseif state == State.TEXT1 then  --"literal"
      if ch == nil then
        self:_raise("literal string opened but not closed.", ln, col)
      end

      if ch.char == '"' then
        state = State.END
      else
        lit = lit .. char(ch)
      end
    elseif state == State.TEXT3_START then
      state = State.TEXT3
      if ch.char ~= "\n" then goto START end
    elseif state == State.TEXT3 then  --"""literal"""
      if ch == nil then
        self:_raise("literal string opened but not closed.", ln, col)
      end

      if ch.char == '"' then
        state = State.END1
      else
        if ch.col < col and ch.char == " " then
          if lit ~= "" then
            if lit:sub(#lit, #lit) ~= "\n" then
              lit = lit .. char(ch)
            end
          end
        else
          lit = lit .. char(ch)
        end
      end
    elseif state == State.END1 then
      if ch.char == '"' then
        state = State.END2
      else
        lit = lit .. '"' .. char(ch)
        state = State.TEXT3
      end
    else  --State.END2
      if ch.char == '"' then
        state = State.END

        if lit:sub(-1) == "\n" then
          lit = lit:sub(1, -2)
        end
      else
        lit = lit .. '""' .. char(ch)
        state = State.TEXT3
      end
    end
  end

  --(3) return
  return Literal.new(ln, col, tmpl and LiteralType.TMPL_STRING or LiteralType.STRING, lit)
end

--Scan a literal regular expression..
--
--@return Re
function Lexer:_scanLiteralRe()
  local reader = self._.reader
  local ln, col, lit

  --(1) state machine
  local State = {START=1, TEXT=2, END=3}
  local state = State.START

  while state ~= State.END do
    local ch = reader:next()

    if ch == nil then
      break
    end

    if state == State.START then
      ln, col, lit = ch.line, ch.col, ""
      state = State.TEXT
    elseif state == State.TEXT then
      if ch.char == '`' then
        state = State.END
      else
        lit = lit .. ch.char
      end
    end
  end

  if state ~= State.END then
    self:_raise("End of input reached with literal regular expression not ended.")
  end

  --(2) return
  return Re.new(ln, col, lit)
end

--Read the next token being a literal text.
function Lexer:nextLiteralText()
  local tok = self:next(TokenType.LITERAL)

  if type(tok.value) == "string" then
    return tok
  else
    self:_raise("literal text expected.", tok)
  end
end
